import music21
import re
import numpy as np
from .constants import DICT_RELATIVE_CHANGE, DICT_TONALITY, EXTENSION_REPLACER, DEGREE_REGEX



def annotation_to_musiclang(text):
    """
    Parse an annotation file or string to musiclang chords with proper durations

    Parameters
    ----------
    text: str
        Filepath or annotation to parse

    Returns
    -------
    score: Score
        MusicLang score

    """
    from .parser import chords_to_musiclang
    print(text)
    chords = analyze_roman_notation(text)
    score = chords_to_musiclang(chords)
    return score


def analyze_roman_notation(text):
    """
    Given a text or a string representing a roman annotation, extract informations

    Parameters
    ----------
    text: str
        Filepath or annotation string to parse

    Returns
    -------


    """

    text = _first_clean(text)
    analysis = music21.converter.parse(text, format="romanText")
    romans = analysis.recurse().getElementsByClass('RomanNumeral')
    chords = []

    for roman in romans:
        prim, sec, key, mode, duration = _get_roman_full(roman)
        notes = [(n.pitch.pitchClass, n.pitch.octave - 5) for n in roman.notes]
        final_degree, extension, final_key, final_mode = _analyze_one_chord(prim, sec, key, mode)
        # Use it
        chords.append([notes, duration, final_degree, extension, (final_key, final_mode)])

    return chords


def _first_clean(text):
    text = text.replace('%', 'ø')
    return text

def analyze_one_chord(figure, key, mode):
    figure = figure.replace('/4', '4').replace('/3', '').replace('/2', '').replace('/5', '5')
    if mode == 'major':
        figure = figure.replace('III+', 'III(+)')

    res = figure.split('/')
    if len(res) == 1:
        prim, sec, ter = res[0], None, None
    elif len(res) == 2:
        prim, sec, ter = res[0], res[1], None
    elif len(res) == 3:
        prim, sec, ter = res[0], res[1], res[2]
    else:
        raise Exception(f'Issue with figure {figure}')

    return _analyze_one_chord(prim, sec, ter, key, mode)

def _analyze_one_chord(prim, sec, ter, key, mode):
    mode = {'major': 'M', 'minor': 'm', 'M': 'M', 'm': 'm'}[mode]
    prim = _clean(prim)
    prim = _replace_special_cases(prim)
    degree, extension = _get_degree_and_extension(prim)
    sec_degree, _ = _get_degree_and_extension(sec)
    ter_degree, _ = _get_degree_and_extension(ter)
    extension = _clean_extension(extension)
    # Get the key, mode of the sec_degree
    final_degree, final_key, final_mode = _get_degree_and_tonality(sec_degree, ter_degree, degree, key, mode)
    return final_degree, extension, final_key, final_mode

def _replace_special_cases(prim):

    #Ge* = ivo*[b3]
    #It* = #ivo*[b3]
    #N = bII6
    #Fr = II*[bV]
    #N6 = bII6

    # if 'N6' in prim:
    #     prim = prim.replace('N6', 'bII6')
    # elif 'N' in prim:
    #     prim = prim.replace('N', 'bII6')
    if 'Ger' in prim:
        pass
    elif 'It' in prim:
        prim = 'It' + prim.replace('It', '')
    elif 'Fr' in prim:
        prim = 'Fr' + prim.replace('Fr', '')

    return prim

def _get_duration(roman):
    """Get the duration of a chord

    Parameters
    ----------
    roman :


    Returns
    -------
    type
        result, fractions.Fraction with denominator limited to 8

    """
    from fractions import Fraction as frac
    return frac(roman.duration.quarterLength).limit_denominator(8)


def _get_roman_full(roman):
    secondary = None
    if roman.secondaryRomanNumeral is not None:
        secondary = roman.secondaryRomanNumeral.primaryFigure
    duration = _get_duration(roman)
    return roman.primaryFigure, secondary, roman.key.tonic.pitchClass, roman.key.mode, duration

def _get_degree_and_extension(data):
    if data is None:
        return None, None
    else:
        degree = DEGREE_REGEX.match(data)[0]
        extension = data.replace(degree, '')
        if degree == 'N' and extension == '':
            extension = '6'
        elif degree == "Ger" and extension == '':
            extension = '7'
        elif degree == "Ger" and extension == '6':
            extension = '65'
        elif degree == "It" and extension == '':
            extension = "7[no5]"
        elif degree == "It" and extension == '53':
            extension = "7[no5]"
        elif degree == "It" and extension == '6':
            extension = "65[no5]"
        elif degree == "It" and extension == '64':
            extension = "2[no5]"
        elif degree == "Fr" and extension == '':
            extension = "7[b5]"
        elif degree == "Fr" and extension == '6':
            extension = "65[b5]"
        elif degree == "Fr":
            extension = extension + "[b5]"
        elif degree == "It":
            extension = extension + "[no5]"
        return degree, extension


from musiclang.analyze.constants import DICT_TONALITY_REVERSE

TONALITY_DICT = {
    "c": 0,
    "d": 2,
    "e": 4,
    "f": 5,
    "g": 7,
    "a": 9,
    "b": 11
}


def get_relative_tone(first_degree, first_mode, current_degree, current_mode):
    delta_degree = (current_degree - first_degree) % 12
    figure = DICT_TONALITY_REVERSE[first_mode][delta_degree, current_mode]
    return figure


def convert_harmony(harmony, get_features=False):
    """
    Convert harmony to relative modulations (Using ternary "/" modulations instead of named modulation).
    That way the harmony does not modulate through the piece.

    Parameters
    ----------
    harmony: str
        Harmony to convert

    Returns
    -------
    new_harmony: str

    """
    lines = harmony.split('\n')
    new_harmony = []
    current_degree = 0
    current_mode = 'M'
    first_degree = 0
    nb_chords = 0
    nb_modulations = 0
    nb_time_signature_change = 0
    bar_duration = 4
    time_signature_den = 4
    time_signature_nom = 4
    has_time = False
    bar_numbers = []
    nb_bars = 0
    first_mode = 'M'
    tonality_line = None
    first = True
    for line in lines:
        if line == "":
            continue
        elif not line.startswith('m'):
            if line.lower().startswith("time"):
                if not has_time:
                    time_signature_nom = int(line.split(':')[1].strip().split('/')[0])
                    time_signature_den = int(line.split(':')[1].strip().split('/')[1])
                    bar_duration = time_signature_nom * 4 / time_signature_den
                nb_time_signature_change += 1
                has_time = True
            new_harmony.append(line)
        else:
            elements = line.split(' ')
            new_line = []
            for el in elements:
                if ":" in el:
                    degree = TONALITY_DICT[el[0].lower()]
                    mode = 'm' if el[0] == el[0].lower() else 'M'
                    flats = len([e for e in el[1:] if e in ['b', '-']])
                    sharps = len([e for e in el[1:] if e in ['#', 's']])
                    degree = degree + sharps - flats
                    if first:
                        first_degree = degree
                        first_mode = mode
                        #new_line.append(el)
                        tonality_line = el.replace(':', '')
                    else:
                        nb_modulations += 1
                    current_degree = degree
                    current_mode = mode
                    first = False
                elif el.startswith('m'):
                    bar_number = int(el[1:].replace('.', '', 1))
                    bar_numbers.append(bar_number)
                    nb_bars = max(bar_number, nb_bars)
                    new_line.append(el)
                elif el.startswith('b') and el[1:].replace('.', '', 1).isdigit():
                    new_el = el
                    new_line.append(new_el)
                else:
                    nb_chords += 1
                    if current_degree != first_degree or current_mode != first_mode:
                        # Relative tone
                        new_el = el + "/" + get_relative_tone(first_degree,
                                                              first_mode, current_degree, current_mode)
                    else:
                        new_el = el
                    new_line.append(new_el)

            new_harmony.append(' '.join(new_line))
    if tonality_line is not None:
        new_harmony.insert(0, "Tonality : " + tonality_line)

    if get_features:
        new_harmony = "\n".join(new_harmony)
        data = {
                "raw_annotation": harmony,
                "annotation": new_harmony,
                "nb_chords": float(nb_chords),
                "nb_bars": float(nb_bars),
                "nb_chords_per_bar": float(nb_chords / nb_bars),
                "nb_modulations": float(nb_modulations),
                "nb_modulations_per_bar": float(nb_modulations / nb_bars),
                "nb_time_signature_change": float(nb_time_signature_change),
                "time_signature_nom": int(time_signature_nom),
                "time_signature_den": int(time_signature_den),
                "bar_duration": float(bar_duration),
                "max_bar_jump": float(np.max(np.diff(bar_numbers))) if len(bar_numbers) > 1 else 0,
                "mode": first_mode,
                "degree": int(first_degree)
                }
        return new_harmony, data

    return "\n".join(new_harmony)

def _get_degree_and_tonality(sec_degree, ter_degree, degree, key, mode):
    if ter_degree is not None:
        # Change the key, mode
        key_add, new_mode = DICT_TONALITY[mode][ter_degree]
        key, mode = (key + key_add) % 12, new_mode

    if sec_degree is None:
        new_key, new_mode = key, mode
    else:
        key_add, new_mode = DICT_TONALITY[mode][sec_degree]
        new_key, new_mode = (key + key_add) % 12, new_mode
    degree, new_mode, key_add = DICT_RELATIVE_CHANGE[new_mode][degree]
    new_key = (new_key + key_add) % 12
    return degree, new_key, new_mode


def _clean(data):
    data = data.replace('42', '2').replace('4/3', '43').replace('/3', '').replace('/', '').replace('/4', '4')
    data = data.replace('%', 'ø')
    return data


def _clean_extension(figure):
    figure = figure.replace('[', '').replace('(', '').replace(']', '').replace(')', '')
    figure = figure.replace('66', '6').replace('67', '7').replace('62', '2').replace('643', '43')
    for key, val in EXTENSION_REPLACER:
        figure = figure.replace(key, val)
    return figure





"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

from .write.library import *


import setuptools

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setuptools.setup(
    name="musiclang",
    version="0.11.0",
    author="Florian GARDIN",
    author_email="fgardin.pro@gmail.com",
    description=("A python package for music notation and generation"
                ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        'Documentation': 'https://musiclang.readthedocs.io/en/latest/',
        'Source': 'https://github.com/MusicLang/musiclang/',
        'Tracker': 'https://github.com/MusicLang/musiclang/issues',
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "lxml",
        "mido==1.2.10",
        "music21==8.1.0",
        "numpy",
        "pandas==1.5.3",
        "scikit-learn==1.1.3",
        "scipy",
        "tensorflow",
        "toml==0.10.2",
        "tomli==2.0.1",
        "xmlschema==2.1.1",
        "transformers==4.26.1",
        "tokenizers==0.13.2",
        "partitura==1.2.1",
        "torch"
                      ],
    packages=setuptools.find_packages(include='*'),
    package_data={'musiclang': ['augmented_net/*.hdf5']},
    include_package_data=True,
    python_requires=">=3.6",
)


from musiclang.library import *
from musiclang import Tonality
import time
import numpy as np


def generate_harmony(nb_bars, tonality, arranger, time_signature=(4, 4), temperature=0.02, one_chord_per_bar=False):
    bar_duration = 4 * frac(time_signature[0], time_signature[1])

    def random_melody_note(proba_scale=0.8):
        scale = np.random.random() < proba_scale
        if scale:
            val = np.random.randint(0, 7)
            note = s0.set_duration(bar_duration)
        else:
            val = np.random.randint(0, 12)
            note = h0.set_duration(bar_duration)
        note.val = val
        return note

    random_melody = sum([random_melody_note() for i in range(nb_bars)], None)
    harmony = melody_to_harmony(random_melody, tonality, arranger, time_signature=time_signature, temperature=temperature, one_chord_per_bar=one_chord_per_bar)

    return harmony


def harmony_to_melody(harmony, tonality, enrich_melody=False, proba_enrich_melody=0.5):
    """
    Given a harmony text and a tonality returns a one note per chord melody relative to the tonality
    """
    from musiclang import ScoreFormatter, VoiceLeading
    chords = ScoreFormatter(harmony).parse()

    rules = VoiceLeading.RULES
    score = VoiceLeading(
        # fixed_voices=['piano__0'],
        voicing=[b0.o(-1), b2.o(-1), b0, b1],
        # exclude_rules=[rules.CROSSING, rules.UNISSON]
    )(chords)
    instr = 'piano__3'
    base_chord = (I % Tonality.from_str(tonality))
    melody = sum([chord.score[instr].to_absolute_note(chord).to_standard_note(base_chord) for chord in score], None)

    if enrich_melody:
        melody = auto_enrich_melody(melody, proba_enrich=proba_enrich_melody)


    return melody


ORNEMENTATIONS = ['suspension_prev', 'suspension_prev_repeat', 'interpolate'
                  ]
INTERPOLATE = ['interpolate']


def auto_enrich_melody(melody, proba_enrich=1.0):
    new_melody = None
    for note in melody.notes:
        random_number = np.random.random()
        if random_number < proba_enrich:
            second_random_number = np.random.random()
            if second_random_number < 0.2:
                new_note = note.add_tag(np.random.choice(ORNEMENTATIONS))
            elif note.duration >= 0.5:
                new_note = note.add_tag(np.random.choice(INTERPOLATE))
            else:
                new_note = note.copy()
        else:
            new_note = note.copy()

        new_melody += new_note

    return new_melody


def melody_to_harmony(melody, tonality, arranger, time_signature=(4, 4), temperature=0.02,
                      one_chord_per_bar=False
                      ):
    """
    Given a one note per chord melody, tonality and time signature, returns a harmony
    """

    def get_real_value(time_signature, beat, duration):
        CONVENTION_DICT = {(6, 8): 2, (2, 2): 2}
        nb_beats = CONVENTION_DICT.get(time_signature, duration)
        ratio = frac(nb_beats, 1) / duration

        return float((beat) * ratio) + 1

    from fractions import Fraction as frac
    if not one_chord_per_bar:
        melody_arranger = [[(note, 1)] for note in melody.notes]
    else:
        # Split melody into bars
        from musiclang import Score
        bar_duration = 4 * frac(time_signature[0], time_signature[1])
        nb_bars = melody.duration // bar_duration
        if bar_duration % melody.duration != 0:
            nb_bars += 1

        chord = (I % Tonality.from_str(tonality))
        score_bars = (I % I.M)(melody).project_on_score(
            Score([chord.set_duration(bar_duration) for i in range(nb_bars)]))
        melody_bars = [c.score['piano__0'] for c in score_bars]
        melody_arranger = [[(n, 1 * (idx == 0)) for idx, n in enumerate(bar.notes)] for bar in melody_bars]


    chord_progression = arranger.arrange(melody_arranger, tonality, temperature=temperature)
    # Convert to harmony grid
    bar_duration = 4 * frac(time_signature[0], time_signature[1])
    text = [f"Time Signature : {time_signature[0]}/{time_signature[1]}",
            f"Tonality : {tonality}"]



    if not one_chord_per_bar:
        current_bar_idx = 0
        current_beat = 0
        time = 0
        current_text = f"m{current_bar_idx}"
        for note, chord in zip(melody.notes, chord_progression):
            bar_idx = time // bar_duration
            if bar_idx != current_bar_idx:
                text.append(current_text)
                current_text = f"m{bar_idx}"
                current_bar_idx = bar_idx

            beat = time % bar_duration
            if beat != 0:
                real_beat = get_real_value(time_signature, beat, bar_duration)
                current_text += " b{:1.2f}".format(real_beat)

            current_text += f" {chord}"
            time += note.duration

        text.append(current_text)
    else:
        text += [f"m{idx} {chord}" for idx, chord in enumerate(chord_progression)]

    return "\n".join(text)



def compose_melody_and_harmony(tonality, nb_bars, arranger,
                               time_signature=(4, 4), temperature=0.02,
                               enrich_melody=False,
                               proba_enrich_melody=0.5
                               ):

    harmony = generate_harmony(nb_bars, tonality, arranger, time_signature=time_signature, temperature=temperature)
    new_melody = harmony_to_melody(harmony, tonality)
    new_harmony = melody_to_harmony(new_melody, tonality, arranger, time_signature=time_signature, temperature=temperature)
    if enrich_melody:
        new_melody = auto_enrich_melody(new_melody, proba_enrich=proba_enrich_melody)

    return new_melody, new_harmony


from musiclang.write.constants import *
from fractions import Fraction as frac
from .utils_metric import bjorklund_algorithm


class Metric:
    """
    A class to handle the metric of a song.
    It is composed of a binary array of 0 and 1, a signature, a tatum (smallest beat) and a number of bars
    The array specifies which tatum which be played in the specific metric

    Examples
    --------
    >>> from musiclang.library import Q, s0
    >>> metric = Metric([1, 0, 0, 1], signature=(4, 4), tatum=Q)
    >>> metric.apply_to_melody([s0, s1])
    s0.hd + s1
    """

    STRONG = "STRONG"
    WEAK = "WEAK"
    SYNCOPATION = "SYNCOPATION"
    SIGNATURES = [(4, 4), (3, 4), (2, 4), (6, 4), (3, 8), (6, 8), (9, 8), (12, 8), (2, 2)]

    def __init__(self, array, signature, tatum=S, nb_bars=1):

        if signature not in Metric.SIGNATURES:
            raise ValueError('Signature is not standard, please use a composite metric')

        self.signature = signature
        self.tatum = tatum
        if isinstance(self.tatum, (tuple, list)):
            self.tatum = frac(*self.tatum)
        self.nb_bars = nb_bars
        self.array = array
        if (self.duration / self.tatum) != len(array):
            raise ValueError('Array size is incompatible with signature and tatum')

    @property
    def nb_notes(self):
        return sum(self.array)

    @property
    def duration(self):
        nom, den = self.signature
        return self.nb_bars * nom * frac(4, den)

    @property
    def bar_duration(self):
        nom, den = self.signature
        return nom * frac(4, den)

    def get_beat_durations(self, array=None):
        if array is None:
            array = self.array
        first_has_note = array[0] == 1
        result = []
        curr_dur = self.tatum
        is_note = first_has_note
        for b in array[1:]:
            if b == 0:
                curr_dur += self.tatum
            elif b == 1:
                result.append((is_note, curr_dur))
                curr_dur = self.tatum
                is_note = True
            else:
                result.append((is_note, curr_dur))
                curr_dur = self.tatum
                is_note = False

        result.append((is_note, curr_dur))

        return result, first_has_note

    def beat_type(self, time):
        """
        Return "STRONG" if the beat is a strong beat
        "WEAK" if weak beat and "SYNCOPATION" otherwise

        Parameters
        ----------
        time

        Returns
        -------

        """
        time = time % self.bar_duration
        if time < 0 or time > self.duration:
            raise ValueError('Time should be between 0 and bar duration')

        if time / self.tatum != int(time / self.tatum):
            return Metric.SYNCOPATION

        if time == 0:
            return Metric.STRONG

        if self.signature in [(4, 4), (2, 2), (2, 4), (6, 4)]:
            return Metric.STRONG if (time % 2) == 0 else Metric.WEAK
        elif self.signature in [(3, 4), (3, 8)]:
            return Metric.STRONG if time == 0 else Metric.WEAK
        elif self.signature in [(3, 8), (2, 8), (1, 8)]:
            return Metric.WEAK
        elif self.signature in [(6, 8), (9, 8), (12, 8)]:
            return Metric.STRONG if (time % frac(3, 2)) == 0 else Metric.WEAK


    @classmethod
    def Euclidian(cls, pulses, signature, tatum, nb_bars=1):
        """
        Following this paper : The Euclidean Algorithm Generates Traditional Musical Rhythms
        http://cgm.cs.mcgill.ca/~godfried/publications/banff.pdf

        Parameters
        ----------
        pulses: Number of pulses in the bar
        signature:
        tatum:

        Returns
        -------
        metric: Metric
                Resulting metric of the song
        """

        steps = cls._nb_steps(signature, tatum, nb_bars)
        return Metric(bjorklund_algorithm(steps, pulses), signature, tatum, nb_bars=nb_bars)




    @classmethod
    def _nb_steps(self, signature, tatum, nb_bars):
        nb_base, den = signature
        base = frac(4, den)
        duration = nb_base * base
        steps = int(nb_bars * frac(duration, tatum))
        return steps

    def euclidian(self, pulses):
        """
        Change the array of this specific metric using Metric.Euclidian
        Following this paper : The Euclidean Algorithm Generates Traditional Musical Rhythms
        http://cgm.cs.mcgill.ca/~godfried/publications/banff.pdf

        Parameters
        ----------
        pulses: Number of pulses in the bar

        Returns
        -------
        metric: Metric

        """
        return Metric.Euclidian(pulses, signature=self.signature, tatum=self.tatum, nb_bars=self.nb_bars)

    @classmethod
    def Full(cls, signature, tatum, nb_bars=1):
        """
        Create a metric with a pulse on each tatum (An array full of 1)

        Parameters
        ----------
        signature: tuple (len == 2)
        tatum: Fraction

        Returns
        -------
        metric: Metric
                Resulting metric of the song
        """

        steps = cls._nb_steps(signature, tatum, nb_bars)
        return Metric([1 for _ in range(steps)], signature, tatum, nb_bars=nb_bars)

    def full(self):
        return Metric.Full(signature=self.signature, tatum=self.tatum, nb_bars=self.nb_bars)

    @classmethod
    def Alternating(cls, n, signature, tatum, nb_bars=1):
        """
        Create an alternating metric with a pulse each n tatum

        Parameters
        ----------
        signature: tuple (len == 2)
        tatum: Fraction

        Returns
        -------
        metric: Metric
                Resulting metric of the song
        """

        steps = cls._nb_steps(signature, tatum, nb_bars)
        return Metric([1 if (a % n) == 0 else 0 for a in range(steps)], signature, tatum, nb_bars=nb_bars)

    def alternating(self, n):
        return Metric.Alternating(n, signature=self.signature, tatum=self.tatum, nb_bars=self.nb_bars)

    def apply_to_score(self, score):
        """
        Apply the rythm to a given score
        Parameters
        ----------
        score

        Returns
        -------

        """
        pass


    def get_array_between(self, start, end):
        if start is None:
            start = 0
        if end is None:
            end = self.duration
        idx_start = start // self.tatum
        idx_end = end // self.tatum
        return [self.array[idx % len(self.array)] for idx in range(idx_start, idx_end)], start, end

    def apply_to_melody(self, melody, expand=True, start=None, end=None):
        """
        Apply the metric to a given melody.
        If expand is True, the notes of the melody will be used circularly until metric duration is satisfied
        Else if the melody has not enough notes it will add silences

        Parameters
        ----------
        melody: Melody
        expand: boolean, default = True

        Returns
        -------
        melody: Melody

        """
        from musiclang import Note, Melody, Silence
        if isinstance(melody, Note):
            notes = [melody.copy()]
        elif isinstance(melody, (list, tuple)):
            notes = melody
        else:
            notes = melody.notes

        array, start, end = self.get_array_between(start, end)
        if not expand and len(notes) < sum(array):
            notes += [Silence(1)] * (sum(array) - len(notes))

        beat_durations, first_has_note = self.get_beat_durations(array)
        result = self._apply_durations_to_melody(notes, beat_durations, first_has_note=first_has_note, expand=expand)
        if result.duration > (end - start):
            # It means the duration of the chord is not divisible by tatum, must shorter the last note
            delta_tatum = (end - start) % self.tatum
            before = result.copy()
            result.notes[-1] = result.notes[-1].set_duration(result.notes[-1].duration - delta_tatum)
            assert result.duration == (end - start), f'{result.duration} {end - start} {delta_tatum} {result} {before}'

        elif result.duration < (end - start):
            delta_tatum = (end - start) % self.tatum
            result.notes[-1] = result.notes[-1].set_duration(result.notes[-1].duration + delta_tatum)

        return result

    @classmethod
    def _apply_durations_to_melody(cls, notes, beat_durations, first_has_note=True, expand=True):
        from musiclang import Silence, Continuation
        melody = []
        for idx, (has_note, beat) in enumerate(beat_durations):
            if idx == 0 and not first_has_note:
                melody.append(Silence(1).set_duration(beat))
            elif not expand and idx > len(notes):
                break
            elif has_note:
                melody.append(notes[idx % len(notes)].set_duration(beat))
            else:
                melody.append(Silence(1).set_duration(beat))

        from musiclang import Melody
        return Melody(melody)

    def complementary(self):
        """
        Get the complementary rhythm of a given rhythm
        eg : Complementary of [1, 0, 0, 0] = [0, 1, 1, 1]
        Keep the same tatum, nb_bars and signature

        Returns
        -------
        metric: Metric
        """
        return Metric([1 - a for a in self.array], signature=self.signature, tatum=self.tatum, nb_bars=self.nb_bars)

    def reversed(self):
        """
        Get the reversed rhythm of a given rhythm
        eg : Reverse of [1, 0, 0, 0] = [0, 0, 0, 1]
        Keep the same tatum, nb_bars and signature

        Returns
        -------
        metric: Metric
        """
        return Metric(self.array[::-1], signature=self.signature, tatum=self.tatum, nb_bars=self.nb_bars)

    def circular_shift(self, n):
        """
        Get the shifted rhythm of a given rhythm by tatum value
        eg : circular_shift(2) of [1, 0, 0, 0] = [0, 0, 1, 0]
        Keep the same tatum, nb_bars and signature

        Returns
        -------
        metric: Metric
        """
        return Metric(self.array[(-n % len(self.array)):] + self.array[:(-n % len(self.array))],
                      signature=self.signature, tatum=self.tatum, nb_bars=self.nb_bars)

    def canon(self, n):
        """
        Get the shifted rhythm of a given rhythm by tatum value, cut the end to fit the metric.
        eg : canon(2) of [1, 0, 0, 0] = [0, 0, 1, 0]
        Keep the same tatum, nb_bars and signature

        Returns
        -------
        metric: Metric
        """
        array_mask = ([0] * n) + ([1] * (len(self.array) - n))
        return Metric(self.array[(n % len(self.array)):] + self.array[:(n % len(self.array))],
                      signature=self.signature, tatum=self.tatum, nb_bars=self.nb_bars).masked(array_mask)

    @classmethod
    def FromArray(cls, array, signature=(4, 4), tatum=None, nb_bars=1):
        """
        Get a metric from a binary array, if several bar we only support constant tatum
        between bar at the time being
        Parameters
        ----------
        array: list[int]
            List of binary
        signature
        tatum
        nb_bars

        Returns
        -------

        """
        if tatum is None:
            nb_notes = len(array)
            nom, den = signature
            duration = nb_bars * nom * frac(4, den)
            tatum = duration / nb_notes
        return Metric(array, signature=signature, tatum=tatum, nb_bars=nb_bars)


    @classmethod
    def FromMelody(cls, melody, signature=(4, 4), tatum=None, nb_bars=1):
        # Get tatum
        if tatum is None:
            tatum = min([note.duration for note in melody.notes])
        array = []
        for note in melody.notes:
            nb_tatums = note.duration / tatum
            if nb_tatums != int(nb_tatums):
                raise ValueError('Could not find a tatum to determine the metric, use a composite rhythm')
            nb_tatums = int(nb_tatums)

            if note.is_note:
                array += [1] + [0] * (nb_tatums - 1)
            else:
                array += [0] + [0] * (nb_tatums - 1)

        if len(array) // nb_bars != len(array) // nb_bars:
            raise ValueError('Nb bars does not divide the metric array')

        return Metric(array, signature=signature, tatum=tatum, nb_bars=nb_bars)

    def from_melody(self, melody):
        return Metric.FromMelody(melody, signature=self.signature, tatum=self.tatum, nb_bars=self.nb_bars)

    def is_note_playing(self, time):
        idx = time / self.tatum
        if idx != int(idx):
            return False
        else:
            return self.array[idx] == 1

    def project(self, melody):
        """
        Project a melody in the given metric

        Parameters
        ----------
        melody: Melody

        Returns
        -------
        melody: Melody
                Filtered melody

        """
        from musiclang import Note
        from musiclang.transform.composing.project import project_on_rhythm
        rhythm = self.apply_to_melody(Note(0, 0, 0, 1))
        projected_melody = project_on_rhythm(rhythm, melody)
        return projected_melody

    def masked(self, mask):
        """
        Get the masked rhythm of a given rhythm by tatum value
        eg : Mask of [1, 0, 1, 0] with [1, 0, 0, 0] is [1, 0, 0, 0]

        Parameters
        ---------
        mask : Metric | list[int]
               Mask to apply to the rhythm
        Returns
        -------
        metric: Metric
        """
        if isinstance(mask, Metric):
            mask = mask.array

        return Metric([e1 * e2 for e1, e2 in zip(self.array, mask)], signature=self.signature,
                      tatum=self.tatum, nb_bars=self.nb_bars)

    def add(self, mask):
        """
        Add to rhythm together
        eg : Mask of [1, 0, 1, 0] with [1, 0, 0, 0] is [1, 0, 0, 0]

        Parameters
        ---------
        mask : Metric | list[int]
               Mask to apply to the rhythm
        Returns
        -------
        metric: Metric
        """
        if isinstance(mask, Metric):
            mask = mask.array

        return Metric([max(e1, e2) for e1, e2 in zip(self.array, mask)], signature=self.signature,
                      tatum=self.tatum, nb_bars=self.nb_bars)


class CompositeMetric:

    def __init__(self, metrics):
        self.metrics = metrics

    def apply_to_melody(self, melody, expand=True):

        from musiclang import Note, Melody, Silence
        if isinstance(melody, Note):
            melody = Melody([melody.copy()])
        elif isinstance(melody, (list, tuple)):
            melody = Melody(melody).copy()

        notes = melody.notes
        durations = sum([m.get_beat_durations()[0] for m in self.metrics], [])
        return Metric.apply_durations_to_melody(notes, durations, expand=expand)

    def masked(self, masks):
        if isinstance(masks, Metric):
            masks = []
        return CompositeMetric([metric.masked(mask) for mask, metric in zip(masks, self.metrics)])

    def canon(self, masks):
        raise NotImplemented()


from musiclang.library import *

def test_absolute_note_basics():

    chord = (I % I.M)
    assert chord.to_pitch(C5) == 0
    assert chord.to_pitch(Cs5) == 1
    chord = (V % II.b.M)['6']
    assert chord.to_pitch(C5) == 0
    assert chord.to_pitch(Cs5) == 1


def test_absolute_note_no_octave():
    assert C5.o(1) == C6



from musiclang.predict.composer import auto_compose
from musiclang import Score
from musiclang.transform.composing.pattern import Pattern
class OrchestralLayer:

    def __init__(self,
                 orchestra,
                 voicing,
                 instruments,
                 metadata=None,
                 fixed_bass=True,
                 voice_leading=True):

        self.orchestra = orchestra
        self.voicing = voicing
        self.fixed_bass = fixed_bass
        self.voice_leading = voice_leading
        self.instruments = instruments
        self.metadata = metadata if metadata is not None else {}

    @classmethod
    def from_json(cls, data):
        return OrchestralLayer(
            orchestra=data['orchestra'],
            voicing=[Score.from_str(note) for note in data['voicing']],
            instruments=data['instruments'],
            fixed_bass=data['fixed_bass'],
            metadata=data.get('metadata', {}),
            voice_leading=data['voice_leading']
        )

    def to_json(self):
        dictionary = {key: item for key, item in self.__dict__.items()}
        dictionary['voicing'] = [str(note) for note in self.voicing]
        if hasattr(dictionary['orchestra'], 'to_json'):
            dictionary['orchestra'] = dictionary['orchestra'].to_json()
        return dictionary


class MelodicLayer:

    def __init__(self, melody, instrument):
        self.melody = melody
        self.instrument = instrument

    def to_json(self):
        return {
                "melody": [str(note) for note in self.melody],
                "instrument": self.instrument
                }

    @classmethod
    def from_json(self, data):
        return MelodicLayer(melody=sum([Score.from_str(note) for note in data['melody']], None), instrument=data['instrument'])

class GlobalLayer:

    def __init__(self, tonality, harmony, patternator, acc_amp='mf'):
        self.tonality = tonality
        self.harmony = harmony
        self.patternator = patternator
        self.acc_amp = acc_amp

    def to_json(self):
        return self.__dict__

    @classmethod
    def from_json(self, data):
        return GlobalLayer(
            tonality=data['tonality'],
            harmony=data['harmony'],
            patternator=data['patternator'],
            acc_amp=data['acc_amp']
        )


class PartComposer:
    """
    Class that is responsible of merging the layer together to pass it to the auto_composer
    """

    def __init__(self, orchestral_layers, melodic_layers, global_layer):
        self.orchestral_layers = self.init_orchestral_layers(orchestral_layers)
        self.melodic_layers = self.init_melodic_layers(melodic_layers)
        self.global_layer = self.init_global_layers(global_layer)

    @classmethod
    def from_json(cls, data):
        return PartComposer(
            orchestral_layers=[OrchestralLayer.from_json(layer) for layer in data['orchestral_layers']],
            melodic_layers=[MelodicLayer.from_json(layer) for layer in data['melodic_layers']],
            global_layer=GlobalLayer.from_json(data['global_layer'])
        )

    def to_json(self):
        return {"global_layer": self.global_layer.to_json(),
                "melodic_layers": [layer.to_json() for layer in self.melodic_layers],
                "orchestral_layers": [layer.to_json() for layer in self.orchestral_layers]
                }

    def init_orchestral_layers(self, orchestral_layers):
        return [OrchestralLayer(**layer) if isinstance(layer, dict) else layer for layer in orchestral_layers]

    def init_melodic_layers(self, melodic_layers):
        return [MelodicLayer(**layer) if isinstance(layer, dict) else layer for layer in melodic_layers]

    def init_global_layers(self, global_layer):
        return GlobalLayer(**global_layer) if isinstance(global_layer, dict) else global_layer

    @classmethod
    def compose(cls, data):
        return cls.from_json(data)()
    def __call__(self):

        melodies = []
        orchestras = []
        voicings = []
        fixed_basses = []
        voice_leadings = []
        instrumentss = []
        all_instrumentss = []
        solo_instruments = []
        tonality = self.global_layer.tonality
        patternator = self.global_layer.patternator
        harmony = self.global_layer.harmony
        acc_amp = self.global_layer.acc_amp
        harmony = harmony.replace('m0', f'm0 {tonality}:')

        instrument_dict = {}
        for layer in self.orchestral_layers:
            orchestras.append(layer.orchestra)
            voicings.append(layer.voicing)
            fixed_basses.append(layer.fixed_bass)
            real_instruments = []
            voice_leadings.append(layer.voice_leading)

            for instrument in layer.instruments:
                clean_instrument = ''.join(instrument.split('__')[:-1])
                instrument_dict[clean_instrument] = instrument_dict.get(clean_instrument, -1) + 1
                index = instrument_dict[clean_instrument]
                real_instruments.append(instrument + f'__{index}')
            instrumentss.append(real_instruments)
            all_instrumentss = all_instrumentss + real_instruments

        for layer in self.melodic_layers:
            melodies.append(layer.melody)
            instrument = layer.instrument
            clean_instrument = ''.join(instrument.split('__')[:-1])
            instrument_dict[clean_instrument] = instrument_dict.get(clean_instrument, -1) + 1
            index = instrument_dict[clean_instrument]
            real_instrument = instrument + f'__{index}'
            solo_instruments.append(real_instrument)
        return auto_compose(melodies, harmony, orchestras, voicings, patternator,
                 tonality, solo_instruments, instrumentss, acc_amp=acc_amp,
                 fixed_bass=fixed_basses, voice_leading=voice_leadings
                 )



# def auto_compose(melody, harmony, orchestra, voicing, patternator,
#                  tonality, solo_instrument, instruments, acc_amp='mf',
#                  fixed_bass=True, voice_leading=True
#                  ):